// volume ops for for altered fdwt.c and idwt.c

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>


#include <fstream>
#include <map>
#include <vector>
#include <algorithm>
using namespace std;

#include "RS_vol_ops.h"

int fdwtOutputSave = 0; //save the FDWT output data as FDWTtransformed.raw, FDWTtransformed.descr, FDWTtransformed.PGM
int fdwtCalcEntropy = 0; //save the FDWT output data memoryless entropy of all subbands in file FDWTtransformed.H0
int fdwtEntropyPredictor = 0; //number of predictor to apply before computing entropy 0-none, 1-(upper+left)/2, 2-MED

int idwtInputSave = 0; //save the IDWT input data as IDWTuntransformed.raw, IDWTuntransformed.descr, IDWTuntransformed.PGM

int dwtFindOrUsePredictors = 0; // find the best predictors for dwt subbands and apply them (fdwt) or undo decorrelation (idwt)

int fdwtIsPredictorFixed[(MAX_LEVELS * 7) + 1 + 1 ] //  valid if dwtFindOrUsePredictors==1; 0 - find predictor in fdwt using H0 estimation, 1 - use the predictor specified in fdwtFixedPredictor, i.e., the actual predictor and not the actualPredictors index
= { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1 };  // (see:RS_band_find_and_apply_predictor())
int fdwtFixedPredictor[(MAX_LEVELS * 7) + 1 + 1]    
= { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1 };
// above tables are indexed with final dwt subband id's, which are ordered level after level, starting from level 0, ordering of final subbands for not top-level 3D decomposition: HLL, LHL, HHL, LLH, HLH, LHH, HHH; 
// top-level 3D decomposition: LLL, HLL, LHL, HHL, LLH, HLH, LHH, HHH; not top-level 2D decomposition: HL, LH, HH; top-level 2D decomposition: LL, HL, LH, HH; 
// e.g., for 4 levels in X,Y and 2 levels in Z (counting levels from 0): HLL0, LHL0, HHL0, LLH0, HLH0, LHH0, HHH0, HLL1, LHL1, HHL1, LLH1, HLH1, LHH1, HHH1, HL-2, LH-2, HH-2, LL-3, HL-3, LH-3, HH-3.




//#define PREDICTORS_ALL 2     // total # of 2d and 3d predictors
//#define PREDICTORS_2D 2      // predictors to use for 2d coding

//const int actualPredictors[PREDICTORS_ALL] = { 0, 4 }; // numbers of predictors to use, to permit disabling certain predictors without needing to modify RS_predictor3d() (see:RS_band_find_and_apply_predictor())

#define PREDICTORS_ALL 11    // total # of 2d and 3d predictors
#define PREDICTORS_2D 5      // predictors to use for 2d coding

const int actualPredictors[PREDICTORS_ALL] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10 }; // numbers of predictors to use, to permit disabling certain predictors without needing to modify RS_predictor3d() (see:RS_band_find_and_apply_predictor())
//
// predNr: 0  - no prediction;
//         1  - P1d_X : left;
//         2  - P1d_Y : upper;
//         3  - AVG_XY : in XY plane AVG: (left+upper)/2;
//         4  - MED_XY : in XY plane MED: median(left, upper, left+upper-upper_left);
//         5  - P1d_Z : above;
//         6  - AVG_XZ : in XZ plane AVG
//         7  - AVG_YZ : in YZ plane AVG
//         8  - MED_XZ : in XZ plane MED
//         9  - MED_YZ : in YZ plane MED
//         10 - AVG3d : 3d: (left+upper+above)/3;




// for dwt_forward_transform and dwt_inverse_transform
// [level][subbandId] == 1 perform DWT part  (e.g., just prediction or just update) resulting in creation of given subband, == 0 skip this part, 
// if both parts of 1d DWT are skipped (e.g., 3 and 5, 7 and 11), reordering of coefficients is skipped as well in dwt_forward/inv_transform
// no subbandId==0, any val. is ok.
int tranfSubbandDo[MAX_LEVELS][15] = {
	// 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14 
	{ -0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
	{ -1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
	{ -2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
	{ -3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
	{ -4, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
};

const int transfOtherSubbandId[15] =				// for given subbandId contains subbandId of other step (0<->1) of DWT transformation 
// 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14 
{ -1, 2, 1, 5, 6, 3, 4,11,12,13,14, 7, 8, 9,10 };


// max transf. level at which denoising is performed, -1 for no denoising at all, levels are numbered from 0
// if level>maxLevel then denoising is called, but uses NOP method
int maxLevelDenoise = -1; 


// denoising arguments for levels and subbands, used by RS_vol_subband_denoise
// [level][subband][param], row elements: 0-subband id(must match row number), 1-denoising transform type tt, 
//														tt=0 - NOP, 2-6 ignore
//														tt=1 - soften 2-rx, 3-ry, 4-rz, 5-wc, 6-wn
//														tt=2 - median denoise 2-rx, 3-ry, 4-rz, 5-6 ignore
//														tt=3 - RCRS denoise 2-rx, 3-ry, 4-rz, 5-6 ignore
//														tt=4 - RCRS2 denoise 2-rx, 3-ry, 4-rz, 5-6 ignore
//														tt=5 - RCRS3 denoise 2-rx, 3-ry, 4-rz, 5-6 ignore
//														tt=6 - Mutekatsu Ryu denoising, 2-6 ignore
//														tt=11 - save, call DENOISE_BAT and load
//														tt=12 - just save, 
//														tt=13 - just load, 
//														tt=14 - load in fwd and save in inverse, 
//														tt=11-14 - 2-saveOrLoad_SaveDescription - if saving, save volume description to .descr file?
//																   3-saveOrLoad_SavePGM - if saving, save 1st slice as 8 - bit PGM also(1 = yes) ?
//																   4-saveOrLoad_LoadWeight - if loading, load full voxel (=100) or only a certain percent of its value (-5 means -5% loaded 105% original)?
//																   5-6 ignore
//														tt=11-14 should work for float 5x3 and int 5x3, others for 5x3 int only, none for 9x7
//
//														for details see RS_vol_subband_denoise()
//
//														tt=100-101 - skip the step, or both steps, see bottom of initialize_RS_vol_ops
int transfParams[MAX_LEVELS][15][7] = {
		// level 0
	   {{ -1, -1, -1, -1, -1, -1, -1 }, // no id==0
		{ 1,  0, 0, 0, 0, 0, 0 },
		{ 2,  0, 0, 0, 0, 0, 0 },
		{ 3,  0, 0, 0, 0, 0, 0 }, //2d
		{ 4,  0, 0, 0, 0, 0, 0 },
		{ 5,  0, 0, 0, 0, 0, 0 }, //2d
		{ 6,  0, 0, 0, 0, 0, 0 },
		{ 7,  0, 0, 0, 0, 0, 0 }, //2d
		{ 8,  0, 0, 0, 0, 0, 0 }, //2d
		{ 9,  0, 0, 0, 0, 0, 0 },
		{ 10, 0, 0, 0, 0, 0, 0 },
		{ 11, 0, 0, 0, 0, 0, 0 }, //2d
		{ 12, 0, 0, 0, 0, 0, 0 }, //2d
		{ 13, 0, 0, 0, 0, 0, 0 },
		{ 14, 0, 0, 0, 0, 0, 0 },},
		// level 1
		{ { -1, -1, -1, -1, -1, -1, -1 }, // no id==0
		{ 1,  0, 0, 0, 0, 0, 0 },
		{ 2,  0, 0, 0, 0, 0, 0 },
		{ 3,  0, 0, 0, 0, 0, 0 }, //2d
		{ 4,  0, 0, 0, 0, 0, 0 },
		{ 5,  0, 0, 0, 0, 0, 0 }, //2d
		{ 6,  0, 0, 0, 0, 0, 0 },
		{ 7,  0, 0, 0, 0, 0, 0 }, //2d
		{ 8,  0, 0, 0, 0, 0, 0 }, //2d
		{ 9,  0, 0, 0, 0, 0, 0 },
		{ 10, 0, 0, 0, 0, 0, 0 },
		{ 11, 0, 0, 0, 0, 0, 0 }, //2d
		{ 12, 0, 0, 0, 0, 0, 0 }, //2d
		{ 13, 0, 0, 0, 0, 0, 0 },
		{ 14, 0, 0, 0, 0, 0, 0 }, },
	// level 2
		{ { -1, -1, -1, -1, -1, -1, -1 }, // no id==0
		{ 1,  0, 0, 0, 0, 0, 0 },
		{ 2,  0, 0, 0, 0, 0, 0 },
		{ 3,  0, 0, 0, 0, 0, 0 }, //2d
		{ 4,  0, 0, 0, 0, 0, 0 },
		{ 5,  0, 0, 0, 0, 0, 0 }, //2d
		{ 6,  0, 0, 0, 0, 0, 0 },
		{ 7,  0, 0, 0, 0, 0, 0 }, //2d
		{ 8,  0, 0, 0, 0, 0, 0 }, //2d
		{ 9,  0, 0, 0, 0, 0, 0 },
		{ 10, 0, 0, 0, 0, 0, 0 },
		{ 11, 0, 0, 0, 0, 0, 0 }, //2d
		{ 12, 0, 0, 0, 0, 0, 0 }, //2d
		{ 13, 0, 0, 0, 0, 0, 0 },
		{ 14, 0, 0, 0, 0, 0, 0 }, },
	// level 3
		{ { -1, -1, -1, -1, -1, -1, -1 }, // no id==0
		{ 1,  0, 0, 0, 0, 0, 0 },
		{ 2,  0, 0, 0, 0, 0, 0 },
		{ 3,  0, 0, 0, 0, 0, 0 }, //2d
		{ 4,  0, 0, 0, 0, 0, 0 },
		{ 5,  0, 0, 0, 0, 0, 0 }, //2d
		{ 6,  0, 0, 0, 0, 0, 0 },
		{ 7,  0, 0, 0, 0, 0, 0 }, //2d
		{ 8,  0, 0, 0, 0, 0, 0 }, //2d
		{ 9,  0, 0, 0, 0, 0, 0 },
		{ 10, 0, 0, 0, 0, 0, 0 },
		{ 11, 0, 0, 0, 0, 0, 0 }, //2d
		{ 12, 0, 0, 0, 0, 0, 0 }, //2d
		{ 13, 0, 0, 0, 0, 0, 0 },
		{ 14, 0, 0, 0, 0, 0, 0 }, },
	// level 4
		{ { -1, -1, -1, -1, -1, -1, -1 }, // no id==0
		{ 1,  0, 0, 0, 0, 0, 0 },
		{ 2,  0, 0, 0, 0, 0, 0 },
		{ 3,  0, 0, 0, 0, 0, 0 }, //2d
		{ 4,  0, 0, 0, 0, 0, 0 },
		{ 5,  0, 0, 0, 0, 0, 0 }, //2d
		{ 6,  0, 0, 0, 0, 0, 0 },
		{ 7,  0, 0, 0, 0, 0, 0 }, //2d
		{ 8,  0, 0, 0, 0, 0, 0 }, //2d
		{ 9,  0, 0, 0, 0, 0, 0 },
		{ 10, 0, 0, 0, 0, 0, 0 },
		{ 11, 0, 0, 0, 0, 0, 0 }, //2d
		{ 12, 0, 0, 0, 0, 0, 0 }, //2d
		{ 13, 0, 0, 0, 0, 0, 0 },
		{ 14, 0, 0, 0, 0, 0, 0 }, },
};





// initialize the module (e.g. load above data from file), to be called in main() before calling compression/decompression
void initialize_RS_vol_ops()
{
	FILE *f;
	int cmd=9999, level, subband, par, unfinished=1, read_err=0, nlevels, i;
                                                                                    // load configuration from TRANSFPAR_DAT
	f = fopen(TRANSFPAR_DAT, "rt");

	if (!f)
		RS_err_exit("\nRS:initialize_RS_vol_ops:cannot open " TRANSFPAR_DAT);

	while (unfinished)
	{
		if (1 != fscanf(f, "%d", &cmd)) read_err = 1;
		switch (cmd)
		{

		case 1001:														// read maxLevelDenoise
			if (1 != fscanf(f, "%d", &maxLevelDenoise)) read_err = 1;
			break;

		case 1002:														// read fdwtOutputSave
			if (1 != fscanf(f, "%d", &fdwtOutputSave)) read_err = 1;
			break;

		case 1003:														// read idwtInputSave
			if (1 != fscanf(f, "%d", &idwtInputSave)) read_err = 1;
			break;

		case 1004:														// read fdwtCalcEntropy
			if (1 != fscanf(f, "%d", &fdwtCalcEntropy)) read_err = 1;
			break;

		case 1005:														// read fdwtEntropyPredictor
			if (1 != fscanf(f, "%d", &fdwtEntropyPredictor)) read_err = 1;
			break;

		case 1010:														// read transfParams
			for (level = 0; level < MAX_LEVELS; level++)
				for (subband = 0; subband < 15; subband++)
					for (par = 0; par < 7; par++)
						if (1 != fscanf(f, "%d", &transfParams[level][subband][par])) read_err = 1;
			break;

		case 1011:														// read 1st page of transfParams, copy it to others
			for (subband = 0; subband < 15; subband++)
				for (par = 0; par < 7; par++)
					if (1 != fscanf(f, "%d", &transfParams[0][subband][par])) read_err = 1;
			for (level = 1; level < MAX_LEVELS; level++)
				for (subband = 0; subband < 15; subband++)
					for (par = 0; par < 7; par++)
						transfParams[level][subband][par] = transfParams[0][subband][par];
			break;

		case 1012:														// read given number of transfParams pages
			if (1 != fscanf(f, "%d", &nlevels)) read_err = 1; 
			if (nlevels>MAX_LEVELS) { read_err = 1; break; }
			for (level = 0; level < nlevels; level++)
				for (subband = 0; subband < 15; subband++)
					for (par = 0; par < 7; par++)
						if (1 != fscanf(f, "%d", &transfParams[level][subband][par])) read_err = 1;
			break;

		case 1020:														// read tranfSubbandDo
			for (level = 0; level < MAX_LEVELS; level++)
				for (subband = 0; subband < 15; subband++)
					if (1 != fscanf(f, "%d", &tranfSubbandDo[level][subband])) read_err = 1;
			break;

		case 1021:														// read 1st row of tranfSubbandDo, copy it to others
				for (subband = 0; subband < 15; subband++)
					if (1 != fscanf(f, "%d", &tranfSubbandDo[0][subband])) read_err = 1;
			for (level = 1; level < MAX_LEVELS; level++)
				for (subband = 0; subband < 15; subband++)
					tranfSubbandDo[level][subband] = tranfSubbandDo[0][subband];
			break;

		case 1022:														// read given number of tranfSubbandDo rows
			if (1 != fscanf(f, "%d", &nlevels)) read_err = 1;
			if (nlevels>MAX_LEVELS) { read_err = 1; break; }
			for (level = 0; level < nlevels; level++)
				for (subband = 0; subband < 15; subband++)
					if (1 != fscanf(f, "%d", &tranfSubbandDo[level][subband])) read_err = 1;
			break;

		case 1030:														// read dwtFindOrUsePredictors
			if (1 != fscanf(f, "%d", &dwtFindOrUsePredictors)) read_err = 1;
			break;

        case 1031:														// read fdwtIsPredictorFixed and fdwtFixedPredictor
            for (i = 0; i< (MAX_LEVELS * 7) + 1; i++)
                if (1 != fscanf(f, "%d", &fdwtIsPredictorFixed[i])) read_err = 1;
            for (i = 0; i< (MAX_LEVELS * 7) + 1; i++)
                if (1 != fscanf(f, "%d", &fdwtFixedPredictor[i])) read_err = 1;
            break;

		case 9999:// all done
			unfinished = 0;
			break;

		default:
			RS_err_exit("\nRS:initialize_RS_vol_ops:bad command in " TRANSFPAR_DAT);
		}
		if (read_err)
			RS_err_exit("\nRS:initialize_RS_vol_ops:bad data in " TRANSFPAR_DAT);
	}

    fclose(f);



																					// partially check loaded data
	//if (maxLevel > MAX_LEVELS)												
	//	RS_err_exit("\nRS:initialize_RS_vol_ops:maxLevel too high.");

	if (fdwtOutputSave != 0 && fdwtOutputSave != 1)
		RS_err_exit("\nRS:initialize_RS_vol_ops: fdwtOutputSave != 0 && fdwtOutputSave != 1 ");
	if (idwtInputSave != 0 && idwtInputSave != 1)
		RS_err_exit("\nRS:initialize_RS_vol_ops: idwtInputSave != 0 && idwtInputSave != 1 ");
	if (fdwtCalcEntropy != 0 && fdwtCalcEntropy != 1)
		RS_err_exit("\nRS:initialize_RS_vol_ops: fdwtCalcEntropy != 0 && fdwtCalcEntropy != 1 ");
	if (dwtFindOrUsePredictors != 0 && dwtFindOrUsePredictors != 1)
		RS_err_exit("\nRS:initialize_RS_vol_ops: dwtFindOrUsePredictors != 0 && dwtFindOrUsePredictors != 1 ");
	for (level = 0; level < MAX_LEVELS; level++)
	{
		for (subband = 1; subband < 15; subband++)
		{
			if (tranfSubbandDo[level][subband] != 0 && tranfSubbandDo[level][subband] != 1)
				RS_err_exit("\nRS:initialize_RS_vol_ops: tranfSubbandDo[level][subband] != 0 && tranfSubbandDo[level][subband] != 1 ");
			if (transfParams[level][subband][0] != subband)
				RS_err_exit("\nRS:initialize_RS_vol_ops: transfParams[level][subband][0] != subband ");
		}
	}
	
    for (i = 0; i < (MAX_LEVELS * 7) + 1; i++)
    {
        int predidx, predfoud = 0;
        if (fdwtIsPredictorFixed[i] && !dwtFindOrUsePredictors)
            RS_err_exit("\nRS:initialize_RS_vol_ops: fdwtIsPredictorFixed[i] && !dwtFindOrUsePredictors ");
        if (fdwtFixedPredictor[i] && !fdwtIsPredictorFixed[i])
            RS_err_exit("\nRS:initialize_RS_vol_ops: fdwtFixedPredictor[i] && !fdwtIsPredictorFixed[i] ");

        for (predidx = 0; predidx < PREDICTORS_ALL; predidx++)
            if (actualPredictors[predidx] == fdwtFixedPredictor[i])
                predfoud = 1;
        if (!predfoud)
            RS_err_exit("\nRS:initialize_RS_vol_ops: !predfoud ");
    }



                                                                                    // denoising modes 100 and 101 fed into transfSubbandDo
	for (level = 0; level < MAX_LEVELS; level++)							
		for (subband = 0; subband < 15; subband++)
		{
			if (transfParams[level][subband][1] == 100 )		//set 0 in transfSubbandDo
				tranfSubbandDo[level][subband] = 0;
			else if (transfParams[level][subband][1] == 101)	//set 0 in transfSubbandDo for both DWT steps
			{
				const int pairSubband = transfOtherSubbandId[subband];
				tranfSubbandDo[level][subband] = 0;
				tranfSubbandDo[level][pairSubband] = 0;
			}
		}

   
	
}


// RS read volume pixel
t_sample RS_vol_get(const t_sample * dataBuffer, const int width, const int height, const int depth, const int x, const int y, const int z)
{
	assert(x >= 0 && x < width);
	assert(y >= 0 && y < height);
	assert(z >= 0 && z < depth);
	return dataBuffer[z*width*height + y*width + x];
}


// RS write volume pixel
void RS_vol_set(const t_sample sample, t_sample * dataBuffer, const int width, const int height, const int depth, const int x, const int y, const int z)
{
	assert(x >= 0 && x < width);
	assert(y >= 0 && y < height);
	assert(z >= 0 && z < depth);
	dataBuffer[z*width*height + y*width + x] = sample;
}


// RS copy pixels between vols, for even X start with even bX and use dX==2, for odd start with odd
void RS_copy_vol(const t_sample * fromBuffer, t_sample * toBuffer, const int width, const int height, const int depth,
	const int bx, const int ex, const int dx,	// band within data buffer - x coordinates range (from, to, step), from==to allowable but copies nothing
	const int by, const int ey, const int dy,	// y
	const int bz, const int ez, const int dz)	// z
{
	int x, y, z;

	assert(bx >= 0 && bx <= ex && ex <= width);
	assert(by >= 0 && by <= ey && ey <= height);
	assert(bz >= 0 && bz <= ez && ez <= depth);

	for (z = bz; z < ez; z += dz) /* slices */
		for (y = by; y < ey; y += dy) /* rows */
			for (x = bx; x < ex; x += dx) /* pixels */
			{
				unsigned bufOffset = z*width*height + y*width + x;
				toBuffer[bufOffset] = fromBuffer[bufOffset];
			}
}


// calculate weighted average of pixels distant from center(x,y,z) by rx, ry and rz, weighting center by wc, others by wo
int RS_vol_band_weighted_average(const t_sample * dataBuffer, const int width, const int height, const int depth, // data buffer
	const int bx, const int ex, const int dx,	// band within data buffer - allowable x coordinates range (from, to, step)
	const int by, const int ey, const int dy,	// y
	const int bz, const int ez, const int dz,	// z
	const int cx, const int cy, const int cz,	// reference voxel position in dataBuffer coordinates
	const int rx, const int ry, const int rz,	// neighbourhood size IN BAND COORINATES
	const int wc, const int wn				    // center and neighbours weights
	)
{
	int x, y, z;	// neighbour  x, y and z, 
	t_sample val;	// placeholder for value,
	int count = 0;
	long long int sum = 0;

	for (z = cz - dz*rz; z <= cz + dz*rz; z += dz)
		for (y = cy - dy*ry; y <= cy + dy*ry; y += dy)
			for (x = cx - dx*rx; x <= cx + dx*rx; x += dx)
				if (x >= bx && x < ex && y >= by && y < ey && z >= bz && z < ez)
				{
					val = RS_vol_get(dataBuffer, width, height, depth, x, y, z);
					if (x == cx && y == cy && z == cz)
					{
						sum += wc*val.i;
						count += wc;
					}
					else
					{
						sum += wn*val.i;
						count += wn;
					}
				}

	assert(count>=0);
	if (sum<0)			// for rounding the average, not truncating
		sum -= count >> 1;
	else
		sum += count >> 1;
	return (int)(sum / count);
}


// calculate median of pixels distant from center(x,y,z) by rx, ry and rz
int RS_vol_band_median(const t_sample * dataBuffer, const int width, const int height, const int depth, // data buffer
	const int bx, const int ex, const int dx,	// band within data buffer - allowable x coordinates range (from, to, step)
	const int by, const int ey, const int dy,	// y
	const int bz, const int ez, const int dz,	// z
	const int cx, const int cy, const int cz,	// reference voxel position in dataBuffer coordinates
	const int rx, const int ry, const int rz	// neighbourhood size IN BAND COORINATES
	)
{
	int x, y, z;	// neighbour  x, y and z, 
	t_sample val;	// placeholder for value,
	int count = 0;
	int median;
	vector <int> tab((rx * 2 + 1)*(ry * 2 + 1)*(rz * 2 + 1));

	for (z = cz - dz*rz; z <= cz + dz*rz; z += dz)
		for (y = cy - dy*ry; y <= cy + dy*ry; y += dy)
			for (x = cx - dx*rx; x <= cx + dx*rx; x += dx)
				if (x >= bx && x < ex && y >= by && y < ey && z >= bz && z < ez)
				{
					val = RS_vol_get(dataBuffer, width, height, depth, x, y, z);
					tab[count++] = val.i;
				}
	assert(count);

	sort(tab.begin(), tab.begin() + count);
	median = tab[count >> 1];
	return median;
}


// RCRS filter - return median if pixel value is outside of range of neighbours, original pixel otherwise
int RS_vol_band_RCRS(const t_sample * dataBuffer, const int width, const int height, const int depth, // data buffer
	const int bx, const int ex, const int dx,	// band within data buffer - allowable x coordinates range (from, to, step)
	const int by, const int ey, const int dy,	// y
	const int bz, const int ez, const int dz,	// z
	const int cx, const int cy, const int cz,	// reference voxel position in dataBuffer coordinates
	const int rx, const int ry, const int rz	// neighbourhood size IN BAND COORINATES
	)
{
	int x, y, z;	// neighbour  x, y and z, 
	t_sample val;	// placeholder for value,
	int count = 0;
	int center = -99999;	// val of central point
	int median;
	int result;
	vector <int> tab((rx * 2 + 1)*(ry * 2 + 1)*(rz * 2 + 1));

	for (z = cz - dz*rz; z <= cz + dz*rz; z += dz)
		for (y = cy - dy*ry; y <= cy + dy*ry; y += dy)
			for (x = cx - dx*rx; x <= cx + dx*rx; x += dx)
				if (x >= bx && x < ex && y >= by && y < ey && z >= bz && z < ez)
				{
					val = RS_vol_get(dataBuffer, width, height, depth, x, y, z);
					tab[count++] = val.i;
					if (x == cx && y == cy && z == cz)
						center = val.i;

				}
	assert(count>1);

	sort(tab.begin(), tab.begin() + count);
	median = tab[count >> 1];

	if (center == tab[0] && center != tab[1] || center == tab[count - 1] && center != tab[count - 2])
		result = median;
	else
		result = center;

	return result;
}



// RCRS 2 filter - ret. closest neighbor if pixel value is outside of range of neighbours, original pixel otherwise
int RS_vol_band_RCRS2(const t_sample * dataBuffer, const int width, const int height, const int depth, // data buffer
	const int bx, const int ex, const int dx,	// band within data buffer - allowable x coordinates range (from, to, step)
	const int by, const int ey, const int dy,	// y
	const int bz, const int ez, const int dz,	// z
	const int cx, const int cy, const int cz,	// reference voxel position in dataBuffer coordinates
	const int rx, const int ry, const int rz	// neighbourhood size IN BAND COORINATES
	)
{
	int x, y, z;	// neighbor  x, y and z, 
	t_sample val;	// placeholder for value,
	int count = 0;
	int center = -99999;	// val of central point
	int result;
	vector <int> tab((rx * 2 + 1)*(ry * 2 + 1)*(rz * 2 + 1));

	for (z = cz - dz*rz; z <= cz + dz*rz; z += dz)
		for (y = cy - dy*ry; y <= cy + dy*ry; y += dy)
			for (x = cx - dx*rx; x <= cx + dx*rx; x += dx)
				if (x >= bx && x < ex && y >= by && y < ey && z >= bz && z < ez)
				{
					val = RS_vol_get(dataBuffer, width, height, depth, x, y, z);
					tab[count++] = val.i;
					if (x == cx && y == cy && z == cz)
						center = val.i;

				}
	assert(count>1);

	sort(tab.begin(), tab.begin() + count);

	if (center == tab[0])
		result = tab[1];
	else if (center == tab[count - 1])
		result = tab[count - 2];
	else
		result = center;

	return result;
}


// RCRS 3 filter - ret. closest neighbor different then pixel value, but not crossing the median of neibourhood
int RS_vol_band_RCRS3(const t_sample * dataBuffer, const int width, const int height, const int depth, // data buffer
	const int bx, const int ex, const int dx,	// band within data buffer - allowable x coordinates range (from, to, step)
	const int by, const int ey, const int dy,	// y
	const int bz, const int ez, const int dz,	// z
	const int cx, const int cy, const int cz,	// reference voxel position in dataBuffer coordinates
	const int rx, const int ry, const int rz	// neighbourhood size IN BAND COORINATES
	)
{
	int x, y, z;	// neighbour  x, y and z, 
	t_sample val;	// placeholder for value,
	int count = 0;
	int center = -99999;	// val of central point
	int result;
	vector <int> tab((rx * 2 + 1)*(ry * 2 + 1)*(rz * 2 + 1));

	for (z = cz - dz*rz; z <= cz + dz*rz; z += dz)
		for (y = cy - dy*ry; y <= cy + dy*ry; y += dy)
			for (x = cx - dx*rx; x <= cx + dx*rx; x += dx)
				if (x >= bx && x < ex && y >= by && y < ey && z >= bz && z < ez)
				{
					val = RS_vol_get(dataBuffer, width, height, depth, x, y, z);
					tab[count++] = val.i;
					if (x == cx && y == cy && z == cz)
						center = val.i;

				}
	assert(count>=4);

	sort(tab.begin(), tab.begin() + count);

	if (center == tab[0])
	{
		int i=1;
		const int fstop = (count & 0x01) ? (count / 2) : (count / 2) - 1 ;

		while (i < fstop && tab[i] == tab[0]) // idac od poczatku znajdz el. inny niz pierwszy, ale nie powyzej polowy wszystkich
			i++;

		result = tab[i];
	}
	else if (center == tab[count - 1])
	{
		int i = count - 2;
		const int bstop = count / 2;

		while (i > bstop && tab[i] == tab[count - 1]) // idac od konca znajdz pierwszy inny niz ostatni, ale nie ponizej polowy wszystkich
			i--;

		result = tab[i];
	}
	else
		result = center;

	return result;
}


//// RS dump vol to stdout
//void RS_dump_vol(char* descr, const t_sample * fromBuffer, const int width, const int height, const int depth)
//{
//	int x, y, z;
//
//	printf("\nvol: %s, width: %d, height: %d, depth: %d", descr, width, height, depth);
//
//	for (z = 0; z < depth; z++) /* slices */
//	{
//		for (y = 0; y < height; y++) /* rows */
//		{
//			printf("\n");
//			for (x = 0; x < width; x++) /* pixels */
//			{
//				printf(" % 3d", RS_vol_get(fromBuffer, width, height, depth, x, y, z));
//			}
//		}
//		printf("\n");
//	}
//}


// save subband, optionally wih description in another file and optionally save first slice as 8-bit pgm 
void RS_save_subband(const t_sample * fromBuffer, const int width, const int height, const int depth, // data buffers
	const int bx, const int ex, const int dx, const int sx,	// subband within data buffer - allowable x coordinates range (from, to, step, starting position)
	const int by, const int ey, const int dy, const int sy,	// y
	const int bz, const int ez, const int dz, const int sz,	// z
	const int fwd, const int level, const int subbandId,
	const char *volName,
	const int saveOrLoad_SaveDescription, const char *descrName,
	const int saveOrLoad_SavePGM, const char * PGMName,
	const int reversible)									
{
	int min, max;
	int x, y, z;
	FILE *f;

	if (reversible)
		min = max = RS_vol_get(fromBuffer, width, height, depth, sx, sy, sz).i;
	else
		min = max = (int) RS_vol_get(fromBuffer, width, height, depth, sx, sy, sz).f;

	f = fopen(volName, "wb");										// save volume as raw
	for (z = sz; z < ez; z += dz) /* For all slices */				// calc min/max and save samples
		for (y = sy; y < ey; y += dy) /* For all rows */
			for (x = sx; x < ex; x += dx) /* For all columns */
			{
				int i;
				t_sample s = RS_vol_get(fromBuffer, width, height, depth, x, y, z);
				if (reversible)
					i = s.i;
				else
					i = (int) s.f;
				if(i < min)
					min = i;
				if (i>max)
					max = i;
				if (1 != fwrite(&s, sizeof(s), 1, f))
					RS_err_exit("\nRS:RS_save_subband problem writing raw file.");
			}
	fclose(f);

	if (saveOrLoad_SaveDescription)
	{
		f = fopen(descrName, "wt");
		fprintf(f,
			"VOLUME (tile) \nwidth: %d, height: %d, depth: %d,\n\n"
			"TRANSF \nlevel: %d, subbandId: %d,\n"
			"fwd: %d, reversible: %d, \n\n"
			"SUBBAND in volume\n"
			"bx: %d, ex: %d, dx: %d, sx: %d,\n"
			"by: %d, ey: %d, dy: %d, sy: %d,\n"
			"bz: %d, ez: %d, dz: %d, sz: %d,\n"
			"width: %d, height: %d, depth: %d,\n"
			"volName: %s, \n\n"
			"SAMPLE (as int) \nmin: %d, max: %d. \n\n",
			width, height, depth,
			level, subbandId,
			fwd, reversible,
			bx, ex, dx, sx,
			by, ey, dy, sy,
			bz, ez, dz, sz,
			(dx - 1 + ex - sx) / dx, (dy - 1 + ey - sy) / dy, (dz - 1 + ez - sz) / dz,
			volName,
			min, max);
		fclose(f);
	}

	if (saveOrLoad_SavePGM)
	{
		f = fopen(PGMName, "wb");										// save first volume slice as 8-bit PGM
		fprintf(f, "P5%c%d %d%c%d%c", 0x0A, (dx - 1 + ex - sx) / dx, (dy - 1 + ey - sy) / dy, 0x0A, 255, 0x0A);//header
		//for (z = sz; z < ez; z += dz) /* For all slices */	commented out - just first xy slice!
		z = sz;
			for (y = sy; y < ey; y += dy) /* For all rows */
				for (x = sx; x < ex; x += dx) /* For all columns */
				{
					int i;
					unsigned char uc;
					t_sample s = RS_vol_get(fromBuffer, width, height, depth, x, y, z);
					if (reversible)
						i = s.i;
					else
						i = (int)s.f;

					if (min != max)
						uc = (unsigned char)(((256 * (long long)(i - min))) / (max - min));
					else
						uc = 0;
					if (1 != fwrite(&uc, 1, 1, f))
						RS_err_exit("\nRS:RS_save_subband problem writing PGM file.");
				}
		fclose(f);
	}
}


void RS_load_subband(t_sample * toBuffer, const int width, const int height, const int depth, // data buffers
	const int bx, const int ex, const int dx, const int sx,	// subband within data buffer - allowable x coordinates range (from, to, step, starting position)
	const int by, const int ey, const int dy, const int sy,	// y
	const int bz, const int ez, const int dz, const int sz,	// z
	const char *volName,
	const int saveOrLoad_LoadWeight,
	const int reversible)
{
	int x, y, z;
	FILE *f;
	int argsused = bx;
	argsused = by;
	argsused = bz;

	f = fopen(volName, "rb");										// save volume as raw
	for (z = sz; z < ez; z += dz) /* For all slices */				// calc min/max and save samples
		for (y = sy; y < ey; y += dy) /* For all rows */
			for (x = sx; x < ex; x += dx) /* For all columns */
			{
				t_sample s;
				
				if (1 != fread(&s, sizeof(s), 1, f))
					RS_err_exit("\nRS:RS_load_subband problem reading raw file.");
				
				if (reversible)
				{
					const long long news = s.i;
					const long long olds = RS_vol_get(toBuffer, width, height, depth, x, y, z).i;

					s.i = (int)((  saveOrLoad_LoadWeight * news + (100 - saveOrLoad_LoadWeight) * olds)/100);
				}
				else
					RS_err_exit("\nRS:RS_load_subband problem: irreversible not implemented.");

				RS_vol_set(s, toBuffer, width, height, depth, x, y, z);
			}
	fclose(f);

}


void RS_level_subband_name(char *name, const int level, const int subbandId, const char * suffix)
{
	sprintf(name, "l%d_s%02d%s", level, subbandId, suffix);
}


// denoises given subband of band of 3d transformation part, or (tt=11..14) performs save/load/external denoise 
// when loading, for load with weight!=100 assumes that part of contents of toBuffer is to be kept (it equals contents of fromBuffer)
void RS_vol_subband_denoise(const t_sample * fromBuffer, t_sample * toBuffer, const int width, const int height, const int depth, // data buffers
	const int bx, const int ex, const int dx, const int sx,	// subband within data buffer - allowable x coordinates range (from, to, step, starting position)
	const int by, const int ey, const int dy, const int sy,	// y
	const int bz, const int ez, const int dz, const int sz,	// z
	const int fwd, const int level, const int subbandId, 	// caled from FDWT? (orherwise IDWT) (inform.), transform level and subband Id (DENOISING step) (1-14)
	const int reversible)									// will reversible DWT be called?
{
	const int 
		tt = level>maxLevelDenoise ? 0 : transfParams[level][subbandId][1],  //if level>maxLevelDenoise then it will copy from fromBuffer to toBuffer
		rx = level>maxLevelDenoise ? 0 : transfParams[level][subbandId][2],
		ry = level>maxLevelDenoise ? 0 : transfParams[level][subbandId][3],
		rz = level>maxLevelDenoise ? 0 : transfParams[level][subbandId][4],
		wc = level>maxLevelDenoise ? 0 : transfParams[level][subbandId][5],
		wn = level>maxLevelDenoise ? 0 : transfParams[level][subbandId][6];

	int x, y, z;

	assert(maxLevelDenoise < MAX_LEVELS);
	assert(level < MAX_LEVELS);
	assert(subbandId >= 1 && subbandId <= 14);
	assert(transfParams[level][subbandId][0] == subbandId);

	if (tt>=11 && tt<=14)	// saving from fromBuffer, loading to toBuffer
	{
		char volName[200],
			descrName[200],
			PGMName[200],
			commandcall[2000];

		const int saveOrLoad_SaveDescription = transfParams[level][subbandId][2];	// if saving, save volume description to .descr file?
		const int saveOrLoad_SavePGM = transfParams[level][subbandId][3];			// if saving, save 1st slice as 8 - bit PGM also(1 = yes) ?
		const int saveOrLoad_LoadWeight = transfParams[level][subbandId][4];		// if loading, load full voxel (=100) or only a certain percent of its value?

		RS_level_subband_name(volName, level, subbandId, ".raw");
		RS_level_subband_name(descrName, level, subbandId, ".descr");
		RS_level_subband_name(PGMName, level, subbandId, ".PGM");

		if ((tt == 14 && !fwd) || (tt == 12) || (tt == 11)) //save
			RS_save_subband(fromBuffer, width, height, depth, bx, ex, dx, sx, by, ey, dy, sy, bz, ez, dz, sz, fwd, level, subbandId,
			volName, saveOrLoad_SaveDescription, descrName, saveOrLoad_SavePGM, PGMName, reversible);

		if (tt == 11)
		{
			sprintf(commandcall, "%s %s %s %d %d %d %d %d %d %d", DENOISE_BAT, volName, descrName, width, height, depth, fwd, level, subbandId, reversible);
			system(commandcall);
		}

		if ((tt == 14 && fwd) || (tt == 13) || (tt == 11))  //load
			RS_load_subband(toBuffer, width, height, depth, bx, ex, dx, sx, by, ey, dy, sy, bz, ez, dz, sz,
			volName, saveOrLoad_LoadWeight, reversible);

		return;
	}

	if (tt == 100 ||	// w initialize_RS_vol_ops przepisywane na 0 do transfSubbandDo, a wiec nie bedzie DWT, wiec wynik odszumienia bedzie nadpisany nieodszumionym
		tt == 101)	    // w initialize_RS_vol_ops przepisywane na 0 do transfSubbandDo dla obu czci (step 0<->1) DWT 
	{
		return;
	}


	if (tt != 0 && !reversible)
		RS_err_exit("\nRS:RS_vol_subband_denoise denoising only for int transforms .");

	//obeys maxLevelDenoise
	for (z = sz; z < ez; z+=dz) /* For all slices */												
		for (y = sy; y < ey; y += dy) /* For all rows */
			for (x = sx; x < ex; x+=dx) /* For all columns */
			{
				t_sample s = RS_vol_get(fromBuffer, width, height, depth, x, y, z);

				if (tt == 0)
				{
				}
				else if (tt == 1)
				{
					s.i = RS_vol_band_weighted_average(fromBuffer, width, height, depth,  //read from dataCopy
						bx, ex, dx, by, ey, dy, bz, ez, dz,
						x, y, z,
						rx, ry, rz,
						wc, wn);
				}
				else if (tt == 2)
				{
					s.i = RS_vol_band_median(fromBuffer, width, height, depth,  //read from dataCopy
						bx, ex, dx, by, ey, dy, bz, ez, dz,
						x, y, z,
						rx, ry, rz);
				}
				else if (tt == 3)
				{
					s.i = RS_vol_band_RCRS(fromBuffer, width, height, depth,  //read from dataCopy
						bx, ex, dx, by, ey, dy, bz, ez, dz,
						x, y, z,
						rx, ry, rz);
				}
				else if (tt == 4)
				{
					s.i = RS_vol_band_RCRS2(fromBuffer, width, height, depth,  //read from dataCopy
						bx, ex, dx, by, ey, dy, bz, ez, dz,
						x, y, z,
						rx, ry, rz);
				}
				else if (tt == 5)
				{
					s.i = RS_vol_band_RCRS3(fromBuffer, width, height, depth,  //read from dataCopy
						bx, ex, dx, by, ey, dy, bz, ez, dz,
						x, y, z,
						rx, ry, rz);
				}
				else if (tt == 6)
				{
					s.i = 0;
				}
				else
					RS_err_exit("\nRS:RS_vol_subband_denoise bad transform type.");

				RS_vol_set(s, toBuffer, width, height, depth, x, y, z);
			}
}




// RS ret. number of L coefficients at the beginning of the non-sparse coefficients range from u0 to u1 (without u1)
int RS_calc_middle(const unsigned int u0, const unsigned int u1)
{
	const unsigned int boundary = u1 - u0;
	return ((u0 % 2) == 0) ? ((boundary + 1) >> 1) : (boundary >> 1);

}


// finds subbandId for given transPart and 1d DWT coordinates
int RS_transPart_to_subbandId(
	const int by, const int ey, // band y boundaries
	const int bz, const int ez, // z
	const int level, const int transPart, 	// transform level and 3D TRANSFORM PART (1-6)
	const int y, const int z, // 3d coordinates y and z of any voxel to be transformed using 1d DWT
	const int decompositionsZ, const int decompositionsY)	// number of decompositions for Z and Y (mind order of them)
{
	int fHz, fHy; // index of the first H sample in z and y dir. == ez/ey if no decomposition in given direction

	assert(transPart >= 1 && transPart <= 6);

	if (transPart == 1 || transPart == 2)
		return transPart;

	if (level < decompositionsZ)  // podpasma dla Z
		fHz = bz + RS_calc_middle(bz, ez);
	else
		fHz = ez;

	if (transPart == 3 || transPart == 4) // kierunek y
	{
		const int subbandBaseId = (transPart == 3 ? 3 : 5);

		if (z >= bz && z<fHz)
			return subbandBaseId;

		if (z >= fHz && z<ez)
			return subbandBaseId + 1;

		assert(0);
	}

	if (level < decompositionsY)  // podpasma dla Y
		fHy = by + RS_calc_middle(by, ey);
	else
		fHy = ey;

	if (transPart == 5 || transPart == 6) // kierunek x
	{
		const int subbandBaseId = (transPart == 5 ? 7 : 11);

		if (z >= bz && z<fHz && y >= by && y<fHy)
			return subbandBaseId;

		if (z >= bz && z<fHz && y >= fHy && y<ey)
			return subbandBaseId + 1;

		if (z >= fHz && z<ez && y >= by && y<fHy)
			return subbandBaseId + 2;

		if (z >= fHz && z<ez && y >= fHy && y<ey)
			return subbandBaseId + 3;

		assert(0);
	}
	assert(0);
	return -1;
}



// denoises band for given 3d transformation part, due to earlier transformations at given level band may have subbands 
void RS_vol_band_denoise_TRANSPART(const t_sample * fromBuffer, t_sample * toBuffer, const int width, const int height, const int depth, // data buffers
	const int bx, const int ex, const int dx, const int sx,	// band within data buffer - allowable x coordinates range (from, to, step, starting position)
	const int by, const int ey, const int dy, const int sy,	// y
	const int bz, const int ez, const int dz, const int sz,	// z
	const int fwd, const int level, const int transPart, 	// caled from FDWT? (orherwise IDWT) (inform.), transform level and 3D TRANSFORM PART (1-6)
	const int decompositionsZ, const int decompositionsY, 	// number of decompositions for Z and Y
	const int reversible)									// will reversible DWT be called?
{
	int fHz, fHy; // index of the first H sample in z and y dir. == ez/ey if no decomposition in given direction

	assert(sx == bx || sx == bx + 1);
	assert(sy == by || sy == by + 1);
	assert(sz == bz || sz == bz + 1);
	assert(transPart >= 1 && transPart <= 6);

	if (transPart == 1 || transPart == 2)
	{
		    RS_vol_subband_denoise(fromBuffer, toBuffer, width, height, depth, bx, ex, dx, sx, by, ey, dy, sy, bz, ez, dz, sz, fwd, level, transPart, reversible);
			assert(transPart   == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, sy   , sz   , decompositionsZ, decompositionsY));
			assert(transPart   == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, ey-1 , ez-1 , decompositionsZ, decompositionsY));
		return;
	}
	
	assert(dz == 1);	// skok o 2 moze byc tylko dla X lub Y
	assert(bz == sz);
	if (level < decompositionsZ)  // podpasma dla Z
		fHz = bz + RS_calc_middle(bz, ez);
	else
		fHz = ez;

	if (transPart == 3 || transPart == 4) // kierunek y
	{
		int subbandId = (transPart == 3 ? 3 : 5);

  		    RS_vol_subband_denoise(fromBuffer, toBuffer, width, height, depth, bx, ex, dx, sx, by, ey, dy, sy, bz, fHz, dz, sz, fwd, level, subbandId, reversible);
			assert(subbandId   == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, sy   , sz   , decompositionsZ, decompositionsY));
			assert(subbandId   == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, ey-1 , fHz-1, decompositionsZ, decompositionsY));

		if (fHz!=ez)
		{
			RS_vol_subband_denoise(fromBuffer, toBuffer, width, height, depth, bx, ex, dx, sx, by, ey, dy, sy, fHz, ez, dz, fHz, fwd, level, subbandId + 1, reversible);
			assert(subbandId+1 == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, sy   , fHz  , decompositionsZ, decompositionsY));
			assert(subbandId+1 == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, ey-1 , ez-1 , decompositionsZ, decompositionsY));
		}

		return;
	}

	assert(dy == 1);	// skok o 2 moze byc tylko dla X 
	assert(by == sy);
	if (level < decompositionsY)  // podpasma dla Y
		fHy = by + RS_calc_middle(by, ey);
	else
		fHy = ey;

	if (transPart == 5 || transPart == 6) // kierunek x
	{
		int subbandId = (transPart == 5 ? 7 : 11);

		    RS_vol_subband_denoise(fromBuffer, toBuffer, width, height, depth, bx, ex, dx, sx, by, fHy, dy, sy, bz, fHz, dz, sz, fwd, level, subbandId, reversible);
			assert(subbandId   == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, sy   , sz   , decompositionsZ, decompositionsY));
			assert(subbandId   == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, fHy-1, fHz-1, decompositionsZ, decompositionsY));

		if (fHy != ey)
		{
			RS_vol_subband_denoise(fromBuffer, toBuffer, width, height, depth, bx, ex, dx, sx, fHy, ey, dy, fHy, bz, fHz, dz, sz, fwd, level, subbandId + 1, reversible);
			assert(subbandId+1 == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, fHy  , sz   , decompositionsZ, decompositionsY));
			assert(subbandId+1 == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, ey-1 , fHz-1, decompositionsZ, decompositionsY));
		}
		if (fHz != ez)
		{
			RS_vol_subband_denoise(fromBuffer, toBuffer, width, height, depth, bx, ex, dx, sx, by, fHy, dy, sy, fHz, ez, dz, fHz, fwd, level, subbandId + 2, reversible);
			assert(subbandId+2 == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, sy   , fHz  , decompositionsZ, decompositionsY));
			assert(subbandId+2 == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, fHy-1, ez-1 , decompositionsZ, decompositionsY));
		}
		if (fHz != ez && fHy != ey)
		{
			RS_vol_subband_denoise(fromBuffer, toBuffer, width, height, depth, bx, ex, dx, sx, fHy, ey, dy, fHy, fHz, ez, dz, fHz, fwd, level, subbandId + 3, reversible);
			assert(subbandId+3 == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, fHy  ,  fHz  , decompositionsZ, decompositionsY));
			assert(subbandId+3 == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, ey-1 ,  ez-1 , decompositionsZ, decompositionsY));
		}

		return;
	}
}

void RS_err_exit(const char * s)  // output error msg and exit
{
	fflush(stdout);
	fflush(stderr);
	fputs(s, stderr);
	fflush(stderr);
	exit(-1);
}




// calculate entropy of a subband being an image only 3x5 reversible, predictors!=0 not supported
double RS_calc_subband_entropy(const t_sample * fromBuffer, const int width, const int height, const int depth, // data buffers
	const int bx, const int ex,
	const int by, const int ey,
	const int bz, const int ez,
	const int predictor)
{
	if (predictor!=0)
		RS_err_exit("RS_calc_subband_entropy: predictor !=0 is obsolete");  // use old version of this function for predictor != 0

	int min, max;
	min = max = RS_vol_get(fromBuffer, width, height, depth, bx, by, bz).i;
	for (int z = bz; z< ez; z++)
		for (int y = by; y< ey; y++)
			for (int x = bx; x < ex; x++)
			{
				const int sample = RS_vol_get(fromBuffer, width, height, depth, x, y, z).i;
				if (sample < min)
					min = sample;
				if (sample > max)
					max = sample;
			}
	const int valRangeSize = max - min + 1;
	int * const tab = (int*)calloc(valRangeSize, sizeof(int));
	if (!tab)
		RS_err_exit("RS_calc_subband_entropy: calloc() fail");  

	for (int z = bz; z < ez; z++)
		for (int y = by; y < ey; y++)
			for (int x = bx; x < ex; x++)
				tab[RS_vol_get(fromBuffer, width, height, depth, x, y, z).i - min]++;
			
	const double invlog2 = 1.0 / log(2.0);  // hepler 
	const double nSamples = double(ex - bx) * double(ey - by) * double(ez - bz); // helper
	double H0 = 0.0;//result
	for (int i = 0; i < valRangeSize; i++)
		if (tab[i])
			H0 -= (double)(tab[i]) * log((double)(tab[i]) / nSamples) * invlog2;

	free(tab);

	return H0;
}


//// calculate entropy of a subband being an image only 3x5 reversible
//double RS_calc_subband_entropy_OLD(const t_sample * fromBuffer, const int width, const int height, const int depth, // data buffers
//	const int bx, const int ex,
//	const int by, const int ey,
//	const int bz, const int ez,
//	const int predictor)
//{
//#define TABSIZE 0x10000
//	int tab[TABSIZE];
//	int i, index;
//
//	const double invlog2 = 1.0 / log(2.0);  // hepler 
//	const int nSamples = (ex - bx) * (ey - by) * (ez - bz); // helper
//	double H0 = 0.0;//result
//
//	for (i = 0; i < TABSIZE; i++)
//		tab[i] = 0;
//
//	for (int z = bz; z< ez; z++)
//		for (int y = by; y< ey; y++)
//			for (int x = bx; x < ex; x++)
//			{
//				const int sample = RS_vol_get(fromBuffer, width, height, depth, x, y, z).i;
//				int predicted = 0;
//
//				if (predictor != 0)
//				{
//					if ((x > bx) && (y > by))
//						switch (predictor)
//						{
//						case 1: predicted = (RS_vol_get(fromBuffer, width, height, depth, x, y - 1, z).i +
//							RS_vol_get(fromBuffer, width, height, depth, x - 1, y, z).i) / 2;
//							break;
//						case 2:
//						{
//							const int
//								a = RS_vol_get(fromBuffer, width, height, depth, x - 1, y, z).i,
//								b = RS_vol_get(fromBuffer, width, height, depth, x, y - 1, z).i,
//								c = RS_vol_get(fromBuffer, width, height, depth, x - 1, y - 1, z).i;
//							int min_ab, max_ab;
//				
//							if (a < b)
//							{
//								min_ab = a;
//								max_ab = b;
//							}
//							else
//							{
//								min_ab = b;
//								max_ab = a;
//							}
//				
//							if (c >= max_ab)
//								predicted = min_ab;
//							else if (c <= min_ab)
//								predicted = max_ab;
//							else
//								predicted = a + b - c;
//						}
//						break;
//						default: RS_err_exit("RS_calc_subband_entropy() bad predictor");
//						}
//					else if ((x == bx) && (y>by))
//						predicted = RS_vol_get(fromBuffer, width, height, depth, x, y - 1, z).i;  // upper
//					else if ((x > bx) && (y == by))
//						predicted = RS_vol_get(fromBuffer, width, height, depth, x - 1, y, z).i;  // left-hand
//					else
//						predicted = 0;
//				}
//
//
//				index = (TABSIZE >> 1) + sample - predicted;
//				if (index<0 || index >= TABSIZE)
//					RS_err_exit("RS_calc_subband_entropy() bad min/max");
//				tab[index]++;
//			}
//
//	for (i = 0; i < TABSIZE; i++)
//		if (tab[i])
//			H0 -= (double)(tab[i]) * log((double)(tab[i]) / (double)(nSamples)) * invlog2;
//	return H0;
//}


//#define ENTROPY_TIMING



// calculate entropy of transformed volume, only 3x5 reversible, as sum of entropies of subbands, 
// using prediction errors of samples (0-none, 1-AVG, 2-MED), and save the entropy to outFileName if it is not NULL
// return entropy
double RS_calc_entropy(t_img_tile_component * const tc, const int predictor, const char *outFileName) 
{
	const int 
		bx = tc->tcx0, 
		by = tc->tcy0, 
		bz = tc->tcz0;
	int ex = tc->tcx1, 
		ey = tc->tcy1, 
		ez = tc->tcz1;
	const int
		maxLevelXY = tc->decompositionsX,
		maxLevelZ = tc->decompositionsZ,
		width = tc->tcx1 - tc->tcx0,
		height = tc->tcy1 - tc->tcy0,
		depth = tc->tcz1 - tc->tcz0;
	const t_sample * const fromBuffer = tc->data;

	int ns; // number of summed subbands
	double H0;	//H0 cumulative
	int lvl;
	FILE *f;

#ifdef ENTROPY_TIMING
	clock_t start = clock();
#endif	

	ns = 0;
	H0 = 0.0;

	if (   !(bx == 0)
		|| !(by == 0)
		|| !(bz == 0)
		|| !(maxLevelXY == tc->decompositionsY)
		|| !(tc->reversible)
		|| !(maxLevelXY>=maxLevelZ)
		) RS_err_exit("RS_calc_entropy() bad call arguments");

	for (lvl = 0; lvl < maxLevelXY; lvl++)   
	{
		const int fHx = RS_calc_middle(bx, ex);
		const int fHy = RS_calc_middle(by, ey);
		const int fHz = lvl < maxLevelZ ? RS_calc_middle(bz, ez) : ez;

		H0 += RS_calc_subband_entropy(fromBuffer, width, height, depth, fHx, ex, by, fHy, bz, fHz, predictor); // HLL
		H0 += RS_calc_subband_entropy(fromBuffer, width, height, depth, bx, fHx, fHy, ey, bz, fHz, predictor); // LHL
		H0 += RS_calc_subband_entropy(fromBuffer, width, height, depth, fHx, ex, fHy, ey, bz, fHz, predictor); // HHL
		ns += 3;

		if(lvl < maxLevelZ)
		{
			H0 += RS_calc_subband_entropy(fromBuffer, width, height, depth, bx, fHx, by, fHy, fHz, ez, predictor); // LLH
			H0 += RS_calc_subband_entropy(fromBuffer, width, height, depth, fHx, ex, by, fHy, fHz, ez, predictor); // HLH
			H0 += RS_calc_subband_entropy(fromBuffer, width, height, depth, bx, fHx, fHy, ey, fHz, ez, predictor); // LHH
			H0 += RS_calc_subband_entropy(fromBuffer, width, height, depth, fHx, ex, fHy, ey, fHz, ez, predictor); // HHH
			ns += 4;
		}

		ex = fHx;
		ey = fHy;
		ez = fHz;
	}
	H0 += RS_calc_subband_entropy(fromBuffer, width, height, depth, bx, ex, by, ey, bz, ez, predictor); // LLL
	ns++;

#ifdef ENTROPY_TIMING
	printf("\nEntropy calculated in %f seconds\n", (float)(clock() - start) / CLOCKS_PER_SEC);
#endif

	if (outFileName)
	{
		f = fopen(outFileName, "wt");
		fprintf(f, "%.0f (H0, bits), predictor: %d subbands: %d\n\n", ceil(H0), predictor, ns);
		fclose(f);
	}

	return H0;
}




/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////// PREDICTORS

inline int min(const int i1, const int i2)
{
	return i1 < i2 ? i1 : i2;
}

inline int max(const int i1, const int i2)
{
	return i1 > i2 ? i1 : i2;
}

inline int median(const int i1, const int i2, const int i3)
{
	return max(min(i1, i2), min(max(i1, i2), i3));
}

inline int predictorAVG(const int left, const int upper)
{
	return (left + upper) >> 1;
}

inline int predictorMED(const int left, const int upper, const int upper_left)
{
	return median(left, upper, left + upper - upper_left);
}

// L left [x-1], U upper [y-1], A above [z-1]
inline int predictorAVG3d(const int A, const int U, const int L)
{
	return (A + U + L) / 3;
}


#define PREDICTORS_IMPLEMENTED 11
static_assert(PREDICTORS_ALL<=PREDICTORS_IMPLEMENTED, "not PREDICTORS_ALL<=PREDICTORS_IMPLEMENTED");

//which predictor to use at the borders of the volume? array works also inside volume
//1st index is the predictor number
//2nd index to be calculated for predicting pixel at x,y,z, when upper top left cornenr of subband is at bx,by,bz: const int borderIdx = ((z == bz) ? 4 : 0) | ((y == by) ? 2 : 0) | ((x == bx) ? 1 : 0);
const int borderPredictors[PREDICTORS_IMPLEMENTED][8]=
{
//z==bz 0  0  0  0  1  1  1  1
//y==by 0  0  1  1  0  0  1  1  
//x==bx 0  1  0  1  0  1  0  1  
	  { 0, 0, 0, 0, 0, 0, 0, 0 }, // predyktory 2d (w czesci, gdzie z==bz)
	  { 1, 2, 1, 5, 1, 2, 1, 0 },
	  { 2, 2, 1, 5, 2, 2, 1, 0 },
	  { 3, 2, 1, 5, 3, 2, 1, 0 },
	  { 4, 2, 1, 5, 4, 2, 1, 0 },
	  { 5, 5, 5, 5, 1, 2, 1, 0 }, // predyktory 3d, 
	  { 6, 5, 6, 5, 1, 2, 1, 0 }, // 2-wymiarowe w 3d, na prostopadlych plaszczyznach bocznych przechodza w 1,2, lub 6
	  { 7, 7, 5, 5, 2, 2, 1, 0 },
	  { 8, 5, 8, 5, 1, 2, 1, 0 },
	  { 9, 9, 5, 5, 2, 2, 1, 0 },
	  {10, 7, 6, 5, 3, 2, 1, 0 }, // pelne 3d, na plaszczyznach bocznych redukuja sie do odpowiednika 2d w danej plaszczyznie
};         

// predNr: 0  - no prediction;
//         1  - P1d_X : left;
//         2  - P1d_Y : upper;
//         3  - AVG_XY : in XY plane AVG: (left+upper)/2;
//         4  - MED_XY : in XY plane MED: median(left, upper, left+upper-upper_left);
//         5  - P1d_Z : above;
//         6  - AVG_XZ : in XZ plane AVG
//         7  - AVG_YZ : in YZ plane AVG
//         8  - MED_XZ : in XZ plane MED
//         9  - MED_YZ : in YZ plane MED
//         10 - AVG3d : 3d: (left+upper+above)/3;
int RS_predictor3d(const t_sample * dataBuffer, const int width, const int height, const int depth, 
	const int x, const int y, const int z, 
	const int predNr)
{
// L left [x-1], U upper [y-1], A above [z-1], UL upper left ...
#define	AU   RS_vol_get(dataBuffer, width, height, depth, x, y - 1, z - 1).i
#define	AL   RS_vol_get(dataBuffer, width, height, depth, x - 1, y, z - 1).i
#define	A    RS_vol_get(dataBuffer, width, height, depth, x, y, z - 1).i
#define	UL   RS_vol_get(dataBuffer, width, height, depth, x - 1, y - 1, z).i
#define	U    RS_vol_get(dataBuffer, width, height, depth, x, y - 1, z).i
#define	L    RS_vol_get(dataBuffer, width, height, depth, x - 1, y, z).i

	switch (predNr)
	{
	case 0: return 0;
	case 1: return L; break;
	case 2: return U; break;
	case 3: return predictorAVG(L, U); break;
	case 4: return predictorMED(L, U, UL); break;
	case 5: return A; break;
	case 6: return predictorAVG(L, A); break;
	case 7: return predictorAVG(U, A); break;
	case 8: return predictorMED(L, A, AL); break;
	case 9: return predictorMED(U, A, AU); break;
	case 10: return predictorAVG3d(A, U, L); break;
	default: RS_err_exit("RS_predictor3d: bad predNr"); return 0;
	}

#undef AU 
#undef AL 
#undef A  
#undef UL 
#undef U  
#undef L  
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////// UNDOING PREDICTION (in-place)


// in-place
void RS_band_undo_apply_predictor(t_sample * buffer, const int width, const int height, const int depth, // data buffer
	const int bx, const int ex, // x start and post-end index
	const int by, const int ey, // y
	const int bz, const int ez, // z
	const int predNr) // predictor number
{
	for (int z = bz; z < ez; z++) 
		for (int y = by; y < ey; y++) 
			for (int x = bx; x < ex; x++) 
			{
				const int borderIdx = ((z == bz) ? 4 : 0) | ((y == by) ? 2 : 0) | ((x == bx) ? 1 : 0);

				t_sample s = RS_vol_get(buffer, width, height, depth, x, y, z); // original sample
				const int p = RS_predictor3d(buffer, width, height, depth, x, y, z, borderPredictors[predNr][borderIdx]); // predicted sample
				s.i += p;
				RS_vol_set(s, buffer, width, height, depth, x, y, z);
			}
}

// tc should be before performing idwt
void RS_undo_apply_predictors(t_img_tile_component * const tc)
{
	if (tc->tcx0 | tc->tcy0 | tc->tcz0)
		RS_err_exit("\nRS:In RS_find_and_apply_predictors (tc->tcx0 | tc->tcy0 | tc->tcz0), not implemented for multiple tiles");
	if (tc->decompositionsX != tc->decompositionsY)
		RS_err_exit("\nRS:In RS_find_and_apply_predictors (tc->decompositionsX != tc->decompositionsY)");
	if (tc->decompositionsZ > tc->decompositionsY)
		RS_err_exit("\nRS:In RS_find_and_apply_predictors (tc->decompositionsZ > tc->decompositionsY)");

	const int
		maxLevel = tc->decompositionsX,
		maxLevelZ = tc->decompositionsZ,
		width = tc->tcx1 - tc->tcx0,
		height = tc->tcy1 - tc->tcy0,
		depth = tc->tcz1 - tc->tcz0;

	FILE * f = fopen(PREDICTORS_DAT, "rt");
	if (!f)
		RS_err_exit("\nRS_undo_apply_predictors: f = fopen(PREDICTORS_DAT, \"wt\") failed");

	int predNr, i1, i2, i3, i4, i5; // fscanf placeolders

	fscanf(f, "Predictors for %d levels in X,Y and %d levels in Z of volume %d %d %d", &i1, &i2, &i3, &i4, &i5);
	if ((i1 != maxLevel) || (i2 != maxLevelZ) || (i3 != width) || (i4 != height) || (i5 != depth))
		RS_err_exit("\nRS_undo_apply_predictors: ((i1 != maxLevel) || (i2 != maxLevelZ) || (i3 != width) || (i4 != height) || (i5 != depth))");

	if (maxLevel == 0)
	{
		predNr = -1; fscanf(f, "%d", &predNr);
		RS_band_undo_apply_predictor(tc->data, width, height, depth, 0, width, 0, height, 0, depth, predNr);
	}
	else
		for (int level = 0; level < maxLevel; ++level)
		{
			const int
				bx = tc->resolution[maxLevel - level].trx0,
				by = tc->resolution[maxLevel - level].try0,
				bz = tc->resolution[maxLevel - level].trz0,
				ex = tc->resolution[maxLevel - level].trx1,
				ey = tc->resolution[maxLevel - level].try1,
				ez = tc->resolution[maxLevel - level].trz1;

			if (bx | by | bz)
				RS_err_exit("\nRS:In RS_find_and_apply_predictors (bx | by | bz)");

			const int // middle in each direction
				mx = RS_calc_middle(0, ex),
				my = RS_calc_middle(0, ey),
				mz = (level < maxLevelZ) ? RS_calc_middle(0, ez) : ez; // number of decompositions in Z may be lower than in other directions

			if (level == maxLevel - 1) // top level?
			{
				predNr = -1; fscanf(f, "%d", &predNr);
				RS_band_undo_apply_predictor(tc->data, width, height, depth, bx, mx, by, my, bz, mz, predNr); //LLL
			}

			predNr = -1; fscanf(f, "%d", &predNr);
			RS_band_undo_apply_predictor(tc->data, width, height, depth, mx, ex, by, my, bz, mz, predNr); //HLL
			predNr = -1; fscanf(f, "%d", &predNr);
			RS_band_undo_apply_predictor(tc->data, width, height, depth, bx, mx, my, ey, bz, mz, predNr); //LHL
			predNr = -1; fscanf(f, "%d", &predNr);
			RS_band_undo_apply_predictor(tc->data, width, height, depth, mx, ex, my, ey, bz, mz, predNr); //HHL

			if (level < maxLevelZ) // did we transform in Z direction?
			{
				predNr = -1; fscanf(f, "%d", &predNr);
				RS_band_undo_apply_predictor(tc->data, width, height, depth, bx, mx, by, my, mz, ez, predNr); //LLH
				predNr = -1; fscanf(f, "%d", &predNr);
				RS_band_undo_apply_predictor(tc->data, width, height, depth, mx, ex, by, my, mz, ez, predNr); //HLH
				predNr = -1; fscanf(f, "%d", &predNr);
				RS_band_undo_apply_predictor(tc->data, width, height, depth, bx, mx, my, ey, mz, ez, predNr); //LHH
				predNr = -1; fscanf(f, "%d", &predNr);
				RS_band_undo_apply_predictor(tc->data, width, height, depth, mx, ex, my, ey, mz, ez, predNr); //HHH
			}
		}

	fscanf(f, "%d", &i1);
	if (i1 != 9999)
		RS_err_exit("\nRS_undo_apply_predictors: bad PREDICTORS_DAT");
	fclose(f);
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////// FINDING PREDICTORS/DOING PREDICTION


// perform prediction with 3d predictor, saving residuum to toBuffer
void RS_band_apply_predictor(const t_sample * fromBuffer, t_sample * toBuffer, const int width, const int height, const int depth, // source and dest. data buffers
	const int bx, const int ex, // x start and post-end index of subband within data buffer
	const int by, const int ey, // y
	const int bz, const int ez, // z
	const int predNr) // predictor number
{
	for (int z = bz; z < ez; z++) /* For all slices*/
		for (int y = by; y < ey; y++) /* For all rows*/
			for (int x = bx; x < ex; x++) /* For all columns*/
			{
				const int borderIdx = ((z == bz) ? 4 : 0) | ((y == by) ? 2 : 0) | ((x == bx) ? 1 : 0);

				t_sample s = RS_vol_get(fromBuffer, width, height, depth, x, y, z); // original sample
				const int p = RS_predictor3d(fromBuffer, width, height, depth, x, y, z, borderPredictors[predNr][borderIdx]); // predicted sample
				s.i -= p;
				RS_vol_set(s, toBuffer, width, height, depth, x, y, z);
			}
}


// find the best predictor and apply it, test smaller # od predictors for 2d image
int RS_band_find_and_apply_predictor(t_sample * buffer, const int width, const int height, const int depth, // data buffer
	const int bx, const int ex, // x start and post-end index
	const int by, const int ey, // y
	const int bz, const int ez, // z
    const int finalSubbandId) // id of this final subband
{
	const int PREDICTORS_TO_TEST = ((bz == 0) && (ez == 1)) ? PREDICTORS_2D : PREDICTORS_ALL;
 
    int bestPred, bestCompressedSize;  
	t_sample * const dataCopy = (t_sample *)malloc(width * height * depth * sizeof(t_sample));
	
    if (fdwtIsPredictorFixed[finalSubbandId])  // 0- find predictor in using H0 estimation, 1- use the predictor fdwtFixedPredictor[finalSubbandId]
        bestPred = fdwtFixedPredictor[finalSubbandId];
    else
        for (int predIdx = 0; predIdx < PREDICTORS_TO_TEST; predIdx++)
        {
            RS_band_apply_predictor(buffer, dataCopy, width, height, depth, bx, ex, by, ey, bz, ez, actualPredictors[predIdx]);

            int cEstSize = int(RS_calc_subband_entropy(dataCopy, width, height, depth, bx, ex, by, ey, bz, ez, 0));
            //if (cEstSize!= int(RS_calc_subband_entropyD(dataCopy, width, height, depth, bx, ex, by, ey, bz, ez, 0)))
            //	RS_err_exit("\nRS:In RS_find_and_apply_predictors entropies differ");


            if ((predIdx == 0) || (cEstSize < bestCompressedSize))
            {
                bestPred = actualPredictors[predIdx];
                bestCompressedSize = cEstSize;
            }
        }

    RS_band_apply_predictor(buffer, dataCopy, width, height, depth, bx, ex, by, ey, bz, ez, bestPred);
	RS_copy_vol(dataCopy, buffer, width, height, depth, bx, ex, 1, by, ey, 1, bz, ez, 1);
	free(dataCopy);

	return bestPred;
} 


// tc should be after performing fdwt
void RS_find_and_apply_predictors(t_img_tile_component * const tc)
{
	if (tc->tcx0 | tc->tcy0 | tc->tcz0)
		RS_err_exit("\nRS:In RS_find_and_apply_predictors (tc->tcx0 | tc->tcy0 | tc->tcz0), not implemented for multiple tiles");
	if (tc->decompositionsX != tc->decompositionsY)
		RS_err_exit("\nRS:In RS_find_and_apply_predictors (tc->decompositionsX != tc->decompositionsY)");
	if (tc->decompositionsZ > tc->decompositionsY)
		RS_err_exit("\nRS:In RS_find_and_apply_predictors (tc->decompositionsZ > tc->decompositionsY)");

	const int
		maxLevel = tc->decompositionsX,
		maxLevelZ = tc->decompositionsZ,
		width = tc->tcx1 - tc->tcx0,
		height = tc->tcy1 - tc->tcy0,
		depth = tc->tcz1 - tc->tcz0;

    int finalSubbandId = 0;

	FILE * f = fopen(PREDICTORS_DAT, "wt");
	if (!f)
		RS_err_exit("\nRS_find_and_apply_predictors: f = fopen(PREDICTORS_DAT, \"wt\") failed");

	fprintf(f, "Predictors for %d levels in X,Y and %d levels in Z of volume %d %d %d", maxLevel, maxLevelZ, width, height, depth);
	fflush(f);

	if (maxLevel == 0)
		fprintf(f, "\n%d\t",
			RS_band_find_and_apply_predictor(tc->data, width, height, depth, 0, width, 0, height, 0, depth, finalSubbandId));
	else
		for (int level = 0; level < maxLevel; ++level)
		{
			const int
				bx = tc->resolution[maxLevel - level].trx0,
				by = tc->resolution[maxLevel - level].try0,
				bz = tc->resolution[maxLevel - level].trz0,
				ex = tc->resolution[maxLevel - level].trx1,
				ey = tc->resolution[maxLevel - level].try1,
				ez = tc->resolution[maxLevel - level].trz1;

			if (bx | by | bz)
				RS_err_exit("\nRS:In RS_find_and_apply_predictors (bx | by | bz)");

			const int // middle in each direction
				mx = RS_calc_middle(0, ex),
				my = RS_calc_middle(0, ey),
				mz = (level < maxLevelZ) ? RS_calc_middle(0, ez) : ez; // number of decompositions in Z may be lower than in other directions

			fprintf(f, "\n");
			if (level == maxLevel - 1) // top level?
			{
				// find and apply predictor to the LLL subband
				fprintf(f, "%d\t",
					RS_band_find_and_apply_predictor(tc->data, width, height, depth, bx, mx, by, my, bz, mz, finalSubbandId++)); //LLL
			}

			// find and apply predictor to the HLL, LHL, and HHL subbands (low Z coordinates)
			fprintf(f, "%d\t",
				RS_band_find_and_apply_predictor(tc->data, width, height, depth, mx, ex, by, my, bz, mz, finalSubbandId++)); //HLL
			fprintf(f, "%d\t",
				RS_band_find_and_apply_predictor(tc->data, width, height, depth, bx, mx, my, ey, bz, mz, finalSubbandId++)); //LHL
			fprintf(f, "%d\t",
				RS_band_find_and_apply_predictor(tc->data, width, height, depth, mx, ex, my, ey, bz, mz, finalSubbandId++)); //HHL

			if (level < maxLevelZ) // did we transform in Z direction?
			{
				// find and apply predictor to the LLH, HLH, LHH, and HHH subbands (high Z coordinates)
				fprintf(f, "%d\t",
					RS_band_find_and_apply_predictor(tc->data, width, height, depth, bx, mx, by, my, mz, ez, finalSubbandId++)); //LLH
				fprintf(f, "%d\t",
					RS_band_find_and_apply_predictor(tc->data, width, height, depth, mx, ex, by, my, mz, ez, finalSubbandId++)); //HLH
				fprintf(f, "%d\t",
					RS_band_find_and_apply_predictor(tc->data, width, height, depth, bx, mx, my, ey, mz, ez, finalSubbandId++)); //LHH
				fprintf(f, "%d\t",
					RS_band_find_and_apply_predictor(tc->data, width, height, depth, mx, ex, my, ey, mz, ez, finalSubbandId++)); //HHH
			}
			fflush(f);
		}                                    

	fprintf(f, "\n\n9999\n");
	fclose(f);
}

