#include "dwt.h"

#define MAX_LEVELS 5

#define TRANSFPAR_DAT "RS__transfpar.dat" // file to load maxLevelDenoise, transfParams and tranfSubbandDo from
#define DENOISE_BAT "RS__denoise.bat" // .bat file to be called for external denoising (transform type = 11)

#define PREDICTORS_DAT "__RS__predictors.dat" // file containing predictors applied before entropy coding of transformed image/volume
											  // created after fdwt if dwtFindOrUsePredictors and read by idwt if dwtFindOrUsePredictors

extern int fdwtOutputSave; //save the FDWT output data as FDWTtransformed.raw, FDWTtransformed.descr, FDWTtransformed.PGM
extern int fdwtCalcEntropy; //save the FDWT output data memoryless entropy of all subbands in file FDWTtransformed.H0
extern int fdwtEntropyPredictor; //number of predictor to apply before computing entropy 0-none, 1-(upper+left)/2, 2-MED

extern int idwtInputSave; //save the IDWT input data as IDWTuntransformed.raw, IDWTuntransformed.descr, IDWTuntransformed.PGM

extern int dwtFindOrUsePredictors; // find best predictors for dwt subbands and apply them (fdwt) or undo decorrelation (idwt)

// volume and helper ops for for altered fdwt.c and idwt.c

//controll data 

extern int tranfSubbandDo[MAX_LEVELS][15];  // for dwt_forward_transform and dwt_inverse_transform [level][subbandId] == 1 perform DWT part resulting in creation of given subband, == 0 skip this part, 

//extern int maxLevelDenoise;  // see RS_vol_ops.c
//extern int transfParams[MAX_LEVELS][15][7]; // see RS_vol_ops.c




// initialize the module (e.g. load above data from file), to be called in main() before calling compression/decompression
void initialize_RS_vol_ops();

//ops

//// RS read volume pixel
//t_sample RS_vol_get(const t_sample * dataBuffer, const int width, const int height, const int depth, const int x, const int y, const int z);
//
//// RS write volume pixel
//void RS_vol_set(const t_sample sample, t_sample * dataBuffer, const int width, const int height, const int depth, const int x, const int y, const int z);


// save subband, optionally wih description in another file and optionally save first slice as 8-bit pgm 
void RS_save_subband(const t_sample * fromBuffer, const int width, const int height, const int depth, // data buffers
	const int bx, const int ex, const int dx, const int sx,	// subband within data buffer - allowable x coordinates range (from, to, step, starting position)
	const int by, const int ey, const int dy, const int sy,	// y
	const int bz, const int ez, const int dz, const int sz,	// z
	const int fwd, const int level, const int subbandId,
	const char *volName,
	const int saveOrLoad_SaveDescription, const char *descrName,
	const int saveOrLoad_SavePGM, const char * PGMName,
	const int reversible);

// RS copy pixels between vols, for even X start with even bX and use dX==2, for odd start with odd
void RS_copy_vol(const t_sample * fromBuffer, t_sample * toBuffer, const int width, const int height, const int depth,
	const int bx, const int ex, const int dx,	// band within data buffer - x coordinates range (from, to, step), from==to allowable but copies nothing
	const int by, const int ey, const int dy,	// y
	const int bz, const int ez, const int dz);	// z

// RS ret. number of L coefficients at the beginning of the non-sparse coefficients range from u0 to u1 (without u1)
int RS_calc_middle(const unsigned int u0, const unsigned int u1);

// finds subbandId for given transPart and 1d DWT coordinates
int RS_transPart_to_subbandId(
	const int by, const int ey, // band y boundaries
	const int bz, const int ez, // z
	const int level, const int transPart, 	// transform level and 3D TRANSFORM PART (1-6)
	const int y, const int z, // 3d coordinates y and z of any voxel to be transformed using 1d DWT
	const int decompositionsZ, const int decompositionsY);	// number of decompositions for Z and Y (mind order of them)

// denoises band for given 3d transformation part, due to earlier transformations at given level band may have subbands 
void RS_vol_band_denoise_TRANSPART(const t_sample * fromBuffer, t_sample * toBuffer, const int width, const int height, const int depth, // data buffers
	const int bx, const int ex, const int dx, const int sx,	// band within data buffer - allowable x coordinates range (from, to, step, starting position)
	const int by, const int ey, const int dy, const int sy,	// y
	const int bz, const int ez, const int dz, const int sz,	// z
	const int fwd, const int level, const int transPart, 	// caled from FDWT? (orherwise IDWT) (inform.), transform level and 3D TRANSFORM PART (1-6)
	const int decompositionsZ, const int decompositionsY, 	// number of decompositions for Z and Y
	const int reversible);									// will reversible DWT be called?

//helper

void RS_err_exit(const char * s); // output error msg and exit

// calculate entropy of transformed volume, only 3x5 reversible, as sum of entropies of subbands, 
// using prediction errors of samples (0-none, 1-AVG, 2-MED), and save the entropy to outFileName if it is not NULL
// return entropy
double RS_calc_entropy(t_img_tile_component * const tc, const int predictor, const char *outFileName); 
																									
// histogram packing and storing original histogram to a file
void RS_histogram_packing(t_sample * buffer, const int width, const int height, const int depth, // data buffer
	const int bx, const int ex, const int dx, const int sx,	// subband within data buffer - allowable x coordinates range (from, to, step, starting position)
	const int by, const int ey, const int dy, const int sy,	// y
	const int bz, const int ez, const int dz, const int sz,	// z
	const char *histogramFileName,
	const int reversible);

// histogram-file based undoing RS_histogram_packing
void RS_histogram_expanding(t_sample * buffer, const int width, const int height, const int depth, // data buffer
	const int bx, const int ex, const int dx, const int sx,	// subband within data buffer - allowable x coordinates range (from, to, step, starting position)
	const int by, const int ey, const int dy, const int sy,	// y
	const int bz, const int ez, const int dz, const int sz,	// z
	const char *histogramFileName,
	const int reversible);

void RS_find_and_apply_predictors(t_img_tile_component * const tc);
void RS_undo_apply_predictors(t_img_tile_component * const tc);
